"""Base dataset methods."""
import os
from mxnet.gluon.data import dataset

# pylint: disable= arguments-differ,unused-argument,missing-docstring,abstract-method

class ClassProperty(object):
    """Readonly @ClassProperty descriptor for internal usage."""
    def __init__(self, fget):
        self.fget = fget

    def __get__(self, owner_self, owner_cls):
        return self.fget(owner_cls)


class VisionDataset(dataset.Dataset):
    """Base Dataset with directory checker.

    Parameters
    ----------
    root : str
        The root path of xxx.names, by default is '~/.mxnet/datasets/foo', where
        `foo` is the name of the dataset.
    """
    def __init__(self, root):
        if not os.path.isdir(os.path.expanduser(root)):
            helper_msg = "{} is not a valid dir. Did you forget to initialize \
                         datasets described in: \
                         `http://gluon-cv.mxnet.io/build/examples_datasets/index.html`? \
                         You need to initialize each dataset only once.".format(root)
            raise OSError(helper_msg)

    @property
    def classes(self):
        raise NotImplementedError

    @property
    def num_class(self):
        """Number of categories."""
        return len(self.classes)


class KeyPointDataset(VisionDataset):
    """Base Dataset for KeyPoint detection.

    Parameters
    ----------
    root : str
        The root path of xxx.names, by defaut is '~/.mxnet/datasets/foo', where
        `foo` is the name of the dataset.
    """
    def __init__(self, root):
        super(KeyPointDataset, self).__init__(root)

    @property
    def num_joints(self):
        """Dataset defined: number of joints provided."""
        return 0

    @property
    def joint_pairs(self):
        """Joint pairs which defines the pairs of joint to be swapped
        when the image is flipped horizontally."""
        return []

    @property
    def parent_joints(self):
        """A dict that defines joint id -> parent_joint_id mapping if applicable, can be empty."""
        return {}



# pylint: disable=wildcard-import
"""Video action recognition, Kinetics400 dataset.
https://deepmind.com/research/open-source/open-source-datasets/kinetics/
"""
from __future__ import absolute_import
from .classification import *



# pylint: disable=missing-docstring,arguments-differ
"""Extended image transformations to video transformations."""
from __future__ import division
import random
import numbers
import numpy as np
from mxnet.gluon import Block

__all__ = ['VideoToTensor', 'VideoNormalize', 'VideoRandomHorizontalFlip', 'VideoMultiScaleCrop',
           'VideoCenterCrop', 'VideoTenCrop', 'VideoGroupTrainTransform', 'VideoGroupValTransform']

class VideoGroupValTransform(Block):
    """Combination of transforms for validation.
        (1) center crop
        (2) to tensor
        (3) normalize
    """

    def __init__(self, size, mean, std, max_intensity=255.0):
        super(VideoGroupValTransform, self).__init__()
        if isinstance(size, numbers.Number):
            self.size = (int(size), int(size))
        else:
            self.size = size
        self.mean = np.asarray(mean).reshape((len(mean), 1, 1))
        self.std = np.asarray(std).reshape((len(std), 1, 1))
        self.max_intensity = max_intensity

    def forward(self, clips):
        h, w, _ = clips[0].shape
        th, tw = self.size
        x1 = int(round((w - tw) / 2.))
        y1 = int(round((h - th) / 2.))

        new_clips = []
        for cur_img in clips:
            crop_img = cur_img[y1:y1+th, x1:x1+tw, :]
            tensor_img = np.transpose(crop_img, axes=(2, 0, 1)) / self.max_intensity
            new_clips.append((tensor_img - self.mean) / self.std)
        return new_clips

class VideoGroupTrainTransform(Block):
    """Combination of transforms for training.
        (1) multiscale crop
        (2) scale
        (3) random horizontal flip
        (4) to tensor
        (5) normalize
    """
    def __init__(self, size, scale_ratios, mean, std, fix_crop=True,
                 more_fix_crop=True, max_distort=1, prob=0.5, max_intensity=255.0):
        super(VideoGroupTrainTransform, self).__init__()

        from ...utils.filesystem import try_import_cv2
        self.cv2 = try_import_cv2()
        self.height = size[0]
        self.width = size[1]
        self.scale_ratios = scale_ratios
        self.fix_crop = fix_crop
        self.more_fix_crop = more_fix_crop
        self.max_distort = max_distort
        self.prob = prob
        self.max_intensity = max_intensity
        self.mean = np.asarray(mean).reshape((len(mean), 1, 1))
        self.std = np.asarray(std).reshape((len(std), 1, 1))

    def fillFixOffset(self, datum_height, datum_width):
        h_off = int((datum_height - self.height) / 4)
        w_off = int((datum_width - self.width) / 4)

        offsets = []
        offsets.append((0, 0))          # upper left
        offsets.append((0, 4*w_off))    # upper right
        offsets.append((4*h_off, 0))    # lower left
        offsets.append((4*h_off, 4*w_off))  # lower right
        offsets.append((2*h_off, 2*w_off))  # center

        if self.more_fix_crop:
            offsets.append((0, 2*w_off))        # top center
            offsets.append((4*h_off, 2*w_off))  # bottom center
            offsets.append((2*h_off, 0))        # left center
            offsets.append((2*h_off, 4*w_off))  # right center

            offsets.append((1*h_off, 1*w_off))  # upper left quarter
            offsets.append((1*h_off, 3*w_off))  # upper right quarter
            offsets.append((3*h_off, 1*w_off))  # lower left quarter
            offsets.append((3*h_off, 3*w_off))  # lower right quarter

        return offsets

    def fillCropSize(self, input_height, input_width):
        crop_sizes = []
        base_size = np.min((input_height, input_width))
        scale_rates = self.scale_ratios
        for h, scale_rate_h in enumerate(scale_rates):
            crop_h = int(base_size * scale_rate_h)
            for w, scale_rate_w in enumerate(scale_rates):
                crop_w = int(base_size * scale_rate_w)
                if (np.absolute(h - w) <= self.max_distort):
                    crop_sizes.append((crop_h, crop_w))

        return crop_sizes

    def forward(self, clips):
        h, w, _ = clips[0].shape

        crop_size_pairs = self.fillCropSize(h, w)
        size_sel = random.randint(0, len(crop_size_pairs)-1)
        crop_height = crop_size_pairs[size_sel][0]
        crop_width = crop_size_pairs[size_sel][1]

        is_flip = random.random() < self.prob
        if self.fix_crop:
            offsets = self.fillFixOffset(h, w)
            off_sel = random.randint(0, len(offsets)-1)
            h_off = offsets[off_sel][0]
            w_off = offsets[off_sel][1]
        else:
            h_off = random.randint(0, h - self.height)
            w_off = random.randint(0, w - self.width)

        new_clips = []
        for cur_img in clips:
            crop_img = cur_img[h_off:h_off+crop_height, w_off:w_off+crop_width, :]
            scale_img = self.cv2.resize(crop_img, (self.width, self.height))
            if is_flip:
                flip_img = np.flip(scale_img, axis=1)
            else:
                flip_img = scale_img
            tensor_img = np.transpose(flip_img, axes=(2, 0, 1)) / self.max_intensity
            new_clips.append((tensor_img - self.mean) / self.std)
        return new_clips

class VideoToTensor(Block):
    """Convert images to tensor.

    Convert a list of images of shape (H x W x C) in the range
    [0, 255] to a float32 tensor of shape (C x H x W) in
    the range [0, 1).

    Parameters
    ----------
    max_intensity : float
        The maximum intensity value to be divided in order to fit the output tensor
        in the range [0, 1).

    Inputs:
        - **data**: a list of frames with shape [H x W x C] and uint8 type

    Outputs:
        - **out**: a list of frames with shape [C x H x W] and float32 type
    """
    def __init__(self, max_intensity=255.0):
        super(VideoToTensor, self).__init__()
        self.max_intensity = max_intensity

    def forward(self, clips):
        new_clips = []
        for cur_img in clips:
            new_clips.append(np.transpose(cur_img, axes=(2, 0, 1)) / self.max_intensity)
        return new_clips

class VideoNormalize(Block):
    """Normalize images with mean and standard deviation.

    Given mean `(m1, ..., mn)` and std `(s1, ..., sn)` for `n` channels,
    this transform normalizes each channel of the input tensor with::

        output[i] = (input[i] - mi) / si

    If mean or std is scalar, the same value will be applied to all channels.

    Parameters
    ----------
    mean : float or tuple of floats
        The mean values.
    std : float or tuple of floats
        The standard deviation values.


    Inputs:
        - **data**: a list of frames with shape [C x H x W]

    Outputs:
        - **out**: a list of normalized frames with shape [C x H x W]
    """

    def __init__(self, mean, std):
        super(VideoNormalize, self).__init__()
        self.mean = np.asarray(mean).reshape((len(mean), 1, 1))
        self.std = np.asarray(std).reshape((len(std), 1, 1))

    def forward(self, clips):
        new_clips = []
        for cur_img in clips:
            new_clips.append((cur_img - self.mean) / self.std)
        return new_clips

class VideoRandomHorizontalFlip(Block):
    """Randomly flip the images left to right with a probability.

    Parameters
    ----------
    prob : float
        The probability value to flip the images.

    Inputs:
        - **data**: a list of frames with shape [H x W x C]

    Outputs:
        - **out**: a list of flipped frames with shape [H x W x C]
    """

    def __init__(self, prob=0.5):
        super(VideoRandomHorizontalFlip, self).__init__()
        self.prob = prob

    def forward(self, clips):
        new_clips = []
        if random.random() < self.prob:
            for cur_img in clips:
                new_clips.append(np.flip(cur_img, axis=1))
        else:
            new_clips = clips
        return new_clips

class VideoMultiScaleCrop(Block):
    """Corner cropping and multi-scale cropping.
    	Two data augmentation techniques introduced in:
        Towards Good Practices for Very Deep Two-Stream ConvNets,
        http://arxiv.org/abs/1507.02159
        Limin Wang, Yuanjun Xiong, Zhe Wang and Yu Qiao

    Parameters:
    ----------
    size : int
    	height and width required by network input, e.g., (224, 224)
    scale_ratios : list
    	efficient scale jittering, e.g., [1.0, 0.875, 0.75, 0.66]
    fix_crop : bool
    	use corner cropping or not. Default: True
    more_fix_crop : bool
    	use more corners or not. Default: True
    max_distort : float
    	maximum distortion. Default: 1

    Inputs:
    	- **data**: a list of frames with shape [H x W x C]

    Outputs:
        - **out**: a list of cropped frames with shape [size x size x C]

    """

    def __init__(self, size, scale_ratios, fix_crop=True,
                 more_fix_crop=True, max_distort=1):
        super(VideoMultiScaleCrop, self).__init__()
        from ...utils.filesystem import try_import_cv2
        self.cv2 = try_import_cv2()
        self.height = size[0]
        self.width = size[1]
        self.scale_ratios = scale_ratios
        self.fix_crop = fix_crop
        self.more_fix_crop = more_fix_crop
        self.max_distort = max_distort

    def fillFixOffset(self, datum_height, datum_width):
        """Fixed cropping strategy

        Inputs:
            - **data**: height and width of input tensor

        Outputs:
            - **out**: a list of locations to crop the image

        """
        h_off = int((datum_height - self.height) / 4)
        w_off = int((datum_width - self.width) / 4)

        offsets = []
        offsets.append((0, 0))          # upper left
        offsets.append((0, 4*w_off))    # upper right
        offsets.append((4*h_off, 0))    # lower left
        offsets.append((4*h_off, 4*w_off))  # lower right
        offsets.append((2*h_off, 2*w_off))  # center

        if self.more_fix_crop:
            offsets.append((0, 2*w_off))        # top center
            offsets.append((4*h_off, 2*w_off))  # bottom center
            offsets.append((2*h_off, 0))        # left center
            offsets.append((2*h_off, 4*w_off))  # right center

            offsets.append((1*h_off, 1*w_off))  # upper left quarter
            offsets.append((1*h_off, 3*w_off))  # upper right quarter
            offsets.append((3*h_off, 1*w_off))  # lower left quarter
            offsets.append((3*h_off, 3*w_off))  # lower right quarter

        return offsets

    def fillCropSize(self, input_height, input_width):
        """Fixed cropping strategy

        Inputs:
            - **data**: height and width of input tensor

        Outputs:
            - **out**: a list of crop sizes to crop the image

        """
        crop_sizes = []
        base_size = np.min((input_height, input_width))
        scale_rates = self.scale_ratios
        for h, scale_rate_h in enumerate(scale_rates):
            crop_h = int(base_size * scale_rate_h)
            for w, scale_rate_w in enumerate(scale_rates):
                crop_w = int(base_size * scale_rate_w)
                if (np.absolute(h - w) <= self.max_distort):
                    crop_sizes.append((crop_h, crop_w))

        return crop_sizes

    def forward(self, clips):
        h, w, _ = clips[0].shape

        crop_size_pairs = self.fillCropSize(h, w)
        size_sel = random.randint(0, len(crop_size_pairs)-1)
        crop_height = crop_size_pairs[size_sel][0]
        crop_width = crop_size_pairs[size_sel][1]

        if self.fix_crop:
            offsets = self.fillFixOffset(h, w)
            off_sel = random.randint(0, len(offsets)-1)
            h_off = offsets[off_sel][0]
            w_off = offsets[off_sel][1]
        else:
            h_off = random.randint(0, h - self.height)
            w_off = random.randint(0, w - self.width)

        new_clips = []
        for cur_img in clips:
            crop_img = cur_img[h_off:h_off+crop_height, w_off:w_off+crop_width, :]
            new_clips.append(self.cv2.resize(crop_img, (self.width, self.height)))
        return new_clips


class VideoCenterCrop(Block):
    """Crop images at the center to have a region of
    the given size. size can be a tuple (target_height, target_width)
    or an integer, in which case the target will be of a square shape (size, size)

    Parameters:
    ----------
    size : int
    	height and width required by network input, e.g., (224, 224)

    Inputs:
    	- **data**: a list of frames with shape [H x W x C]

    Outputs:
        - **out**: a list of cropped frames with shape [size x size x C]
    """

    def __init__(self, size):
        super(VideoCenterCrop, self).__init__()
        if isinstance(size, numbers.Number):
            self.size = (int(size), int(size))
        else:
            self.size = size

    def forward(self, clips):
        h, w, _ = clips[0].shape
        th, tw = self.size
        x1 = int(round((w - tw) / 2.))
        y1 = int(round((h - th) / 2.))

        new_clips = []
        for cur_img in clips:
            new_clips.append(cur_img[y1:y1+th, x1:x1+tw, :])
        return new_clips

class VideoThreeCrop(Block):
    """This method crops 3 regions. All regions will be in shape
    :obj`size`. Depending on the situation, these regions may consist of:
        (1) 1 center, 1 top and 1 bottom
        (2) 1 center, 1 left and 1 right

    Parameters:
    ----------
    size : int
        height and width required by network input, e.g., (224, 224)

    Inputs:
        - **data**: a list of N frames with shape [H x W x C]

    Outputs:
        - **out**: a list of 3xN cropped frames with shape [size x size x C]
    """

    def __init__(self, size):
        super(VideoThreeCrop, self).__init__()
        if isinstance(size, numbers.Number):
            self.size = (int(size), int(size))
        else:
            self.size = size

    def forward(self, clips):
        h, w, _ = clips[0].shape
        th, tw = self.size
        assert th == h or tw == w

        if th == h:
            w_step = (w - tw) // 2
            offsets = []
            offsets.append((0, 0))  # left
            offsets.append((2 * w_step, 0))  # right
            offsets.append((w_step, 0))  # middle
        elif tw == w:
            h_step = (h - th) // 2
            offsets = []
            offsets.append((0, 0))  # top
            offsets.append((0, 2 * h_step))  # down
            offsets.append((0, h_step))  # middle

        new_clips = []
        for ow, oh in offsets:
            for cur_img in clips:
                crop_img = cur_img[oh:oh+th, ow:ow+tw, :]
                new_clips.append(crop_img)
        return new_clips



class VideoTenCrop(Block):
    """Crop 10 regions from images.
    This is performed same as:
    http://chainercv.readthedocs.io/en/stable/reference/transforms.html#ten-crop

    This method crops 10 regions. All regions will be in shape
    :obj`size`. These regions consist of 1 center crop and 4 corner
    crops and horizontal flips of them.
    The crops are ordered in this order.
    * center crop
    * top-left crop
    * bottom-left crop
    * top-right crop
    * bottom-right crop
    * center crop (flipped horizontally)
    * top-left crop (flipped horizontally)
    * bottom-left crop (flipped horizontally)
    * top-right crop (flipped horizontally)
    * bottom-right crop (flipped horizontally)

    Parameters:
    ----------
    size : int
    	height and width required by network input, e.g., (224, 224)

    Inputs:
    	- **data**: a list of N frames with shape [H x W x C]

    Outputs:
        - **out**: a list of 10xN frames with shape [size x size x C]

    """
    def __init__(self, size):
        super(VideoTenCrop, self).__init__()
        if isinstance(size, numbers.Number):
            self.size = (int(size), int(size))
        else:
            self.size = size

    def forward(self, clips):
        h, w, _ = clips[0].shape
        oh, ow = self.size
        if h < oh or w < ow:
            raise ValueError("Cannot crop area {} from image with size \
            	({}, {})".format(str(self.size), h, w))

        new_clips = []
        for cur_img in clips:
            center = cur_img[(h - oh) // 2:(h + oh) // 2, (w - ow) // 2:(w + ow) // 2, :]
            tl = cur_img[0:oh, 0:ow, :]
            bl = cur_img[h - oh:h, 0:ow, :]
            tr = cur_img[0:oh, w - ow:w, :]
            br = cur_img[h - oh:h, w - ow:w, :]
            new_clips.append(center)
            new_clips.append(tl)
            new_clips.append(bl)
            new_clips.append(tr)
            new_clips.append(br)
            new_clips.append(np.flip(center, axis=1))
            new_clips.append(np.flip(tl, axis=1))
            new_clips.append(np.flip(bl, axis=1))
            new_clips.append(np.flip(tr, axis=1))
            new_clips.append(np.flip(br, axis=1))
        return new_clips



# pylint: disable=line-too-long,too-many-lines,missing-docstring,arguments-differ,unused-argument
import mxnet as mx
from mxnet import init
from mxnet.gluon import nn
from mxnet.gluon.nn import HybridBlock
from ...nn.block import Consensus
from ..resnetv1b import resnet18_v1b, resnet34_v1b, resnet50_v1b, resnet101_v1b, resnet152_v1b

__all__ = ['resnet18_v1b_sthsthv2', 'resnet34_v1b_sthsthv2', 'resnet50_v1b_sthsthv2',
           'resnet101_v1b_sthsthv2', 'resnet152_v1b_sthsthv2', 'resnet18_v1b_kinetics400',
           'resnet34_v1b_kinetics400', 'resnet50_v1b_kinetics400', 'resnet101_v1b_kinetics400',
           'resnet152_v1b_kinetics400']

class ActionRecResNetV1b(HybridBlock):
    r"""ResNet models for video action recognition

    Parameters
    ----------
    depth : int, number of layers in a ResNet model
    nclass : int, number of classes
    pretrained_base : bool, load pre-trained weights or not
    dropout_ratio : float, add a dropout layer with a ratio p to prevent overfitting
    init_std : float, initialization of the standard deviation value of
               the output (classification) layer

    Input: a single image
    Output: a single predicted action label
    """
    def __init__(self, depth, nclass, pretrained_base=True,
                 dropout_ratio=0.5, init_std=0.01,
                 partial_bn=False, **kwargs):
        super(ActionRecResNetV1b, self).__init__()

        if depth == 18:
            pretrained_model = resnet18_v1b(pretrained=pretrained_base, **kwargs)
            self.expansion = 1
        elif depth == 34:
            pretrained_model = resnet34_v1b(pretrained=pretrained_base, **kwargs)
            self.expansion = 1
        elif depth == 50:
            pretrained_model = resnet50_v1b(pretrained=pretrained_base, **kwargs)
            self.expansion = 4
        elif depth == 101:
            pretrained_model = resnet101_v1b(pretrained=pretrained_base, **kwargs)
            self.expansion = 4
        elif depth == 152:
            pretrained_model = resnet152_v1b(pretrained=pretrained_base, **kwargs)
            self.expansion = 4
        else:
            print('No such ResNet configuration for depth=%d' % (depth))

        self.dropout_ratio = dropout_ratio
        self.init_std = init_std

        self.conv1 = pretrained_model.conv1
        self.bn1 = pretrained_model.bn1
        self.relu = pretrained_model.relu
        self.maxpool = pretrained_model.maxpool
        self.layer1 = pretrained_model.layer1
        self.layer2 = pretrained_model.layer2
        self.layer3 = pretrained_model.layer3
        self.layer4 = pretrained_model.layer4
        self.avgpool = pretrained_model.avgpool
        self.flat = pretrained_model.flat
        self.drop = nn.Dropout(rate=self.dropout_ratio)
        self.output = nn.Dense(units=nclass, in_units=512 * self.expansion,
                               weight_initializer=init.Normal(sigma=self.init_std))
        self.output.initialize()

    def hybrid_forward(self, F, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)

        x = self.avgpool(x)
        x = self.flat(x)
        x = self.drop(x)
        x = self.output(x)
        return x

class ActionRecResNetV1bTSN(HybridBlock):
    r"""ResNet models with temporal segments for video action recognition

    Parameters
    ----------
    depth : int, number of layers in a ResNet model
    nclass : int, number of classes
    pretrained_base : bool, load pre-trained weights or not
    dropout_ratio : float, add a dropout layer with a ratio p to prevent overfitting
    init_std : float, initialization of the standard deviation value of
               the output (classification) layer

    Input: N images from N segments in a single video
    Output: a single predicted action label
    """
    def __init__(self, depth, nclass, pretrained_base=True,
                 partial_bn=False, num_segments=3,
                 dropout_ratio=0.5, init_std=0.01,
                 **kwargs):
        super(ActionRecResNetV1bTSN, self).__init__()

        self.basenet = ActionRecResNetV1b(depth=depth,
                                          nclass=nclass,
                                          pretrained_base=pretrained_base,
                                          partial_bn=partial_bn)
        self.tsn_consensus = Consensus(nclass=nclass, num_segments=num_segments)

    def hybrid_forward(self, F, x):
        pred = self.basenet(x)
        consensus_out = self.tsn_consensus(pred)
        return consensus_out

def resnet18_v1b_sthsthv2(nclass=174, pretrained=False, tsn=False, partial_bn=False,
                          num_segments=1, root='~/.mxnet/models', ctx=mx.cpu(), **kwargs):
    if tsn:
        model = ActionRecResNetV1bTSN(depth=18,
                                      nclass=nclass,
                                      partial_bn=partial_bn,
                                      num_segments=num_segments)
    else:
        model = ActionRecResNetV1b(depth=18,
                                   nclass=nclass,
                                   partial_bn=partial_bn)

    if pretrained:
        from ..model_store import get_model_file
        model.load_parameters(get_model_file('resnet18_v1b_sthsthv2',
                                             tag=pretrained, root=root))
        from ...data import SomethingSomethingV2Attr
        attrib = SomethingSomethingV2Attr()
        model.classes = attrib.classes
    model.collect_params().reset_ctx(ctx)
    return model

def resnet34_v1b_sthsthv2(nclass=174, pretrained=False, tsn=False, partial_bn=False,
                          num_segments=1, root='~/.mxnet/models', ctx=mx.cpu(), **kwargs):
    if tsn:
        model = ActionRecResNetV1bTSN(depth=34,
                                      nclass=nclass,
                                      partial_bn=partial_bn,
                                      num_segments=num_segments)
    else:
        model = ActionRecResNetV1b(depth=34,
                                   nclass=nclass,
                                   partial_bn=partial_bn)

    if pretrained:
        from ..model_store import get_model_file
        model.load_parameters(get_model_file('resnet34_v1b_sthsthv2',
                                             tag=pretrained, root=root))
        from ...data import SomethingSomethingV2Attr
        attrib = SomethingSomethingV2Attr()
        model.classes = attrib.classes
    model.collect_params().reset_ctx(ctx)
    return model

def resnet50_v1b_sthsthv2(nclass=174, pretrained=False, tsn=False, partial_bn=False,
                          num_segments=1, root='~/.mxnet/models', ctx=mx.cpu(), **kwargs):
    if tsn:
        model = ActionRecResNetV1bTSN(depth=50,
                                      nclass=nclass,
                                      partial_bn=partial_bn,
                                      num_segments=num_segments)
    else:
        model = ActionRecResNetV1b(depth=50,
                                   nclass=nclass,
                                   partial_bn=partial_bn)

    if pretrained:
        from ..model_store import get_model_file
        model.load_parameters(get_model_file('resnet50_v1b_sthsthv2',
                                             tag=pretrained, root=root))
        from ...data import SomethingSomethingV2Attr
        attrib = SomethingSomethingV2Attr()
        model.classes = attrib.classes
    model.collect_params().reset_ctx(ctx)
    return model

def resnet101_v1b_sthsthv2(nclass=174, pretrained=False, tsn=False, partial_bn=False,
                           num_segments=1, root='~/.mxnet/models', ctx=mx.cpu(), **kwargs):
    if tsn:
        model = ActionRecResNetV1bTSN(depth=101,
                                      nclass=nclass,
                                      partial_bn=partial_bn,
                                      num_segments=num_segments)
    else:
        model = ActionRecResNetV1b(depth=101,
                                   nclass=nclass,
                                   partial_bn=partial_bn)

    if pretrained:
        from ..model_store import get_model_file
        model.load_parameters(get_model_file('resnet101_v1b_sthsthv2',
                                             tag=pretrained, root=root))
        from ...data import SomethingSomethingV2Attr
        attrib = SomethingSomethingV2Attr()
        model.classes = attrib.classes
    model.collect_params().reset_ctx(ctx)
    return model

def resnet152_v1b_sthsthv2(nclass=174, pretrained=False, tsn=False, partial_bn=False,
                           num_segments=1, root='~/.mxnet/models', ctx=mx.cpu(), **kwargs):
    if tsn:
        model = ActionRecResNetV1bTSN(depth=152,
                                      nclass=nclass,
                                      partial_bn=partial_bn,
                                      num_segments=num_segments)
    else:
        model = ActionRecResNetV1b(depth=152,
                                   nclass=nclass,
                                   partial_bn=partial_bn)

    if pretrained:
        from ..model_store import get_model_file
        model.load_parameters(get_model_file('resnet152_v1b_sthsthv2',
                                             tag=pretrained, root=root))
        from ...data import SomethingSomethingV2Attr
        attrib = SomethingSomethingV2Attr()
        model.classes = attrib.classes
    model.collect_params().reset_ctx(ctx)
    return model

def resnet18_v1b_kinetics400(nclass=400, pretrained=False, tsn=False, partial_bn=False,
                             num_segments=1, root='~/.mxnet/models', ctx=mx.cpu(), **kwargs):
    if tsn:
        model = ActionRecResNetV1bTSN(depth=18,
                                      nclass=nclass,
                                      partial_bn=partial_bn,
                                      num_segments=num_segments)
    else:
        model = ActionRecResNetV1b(depth=18,
                                   nclass=nclass,
                                   partial_bn=partial_bn)

    if pretrained:
        from ..model_store import get_model_file
        model.load_parameters(get_model_file('resnet18_v1b_kinetics400',
                                             tag=pretrained, root=root))
        from ...data import Kinetics400Attr
        attrib = Kinetics400Attr()
        model.classes = attrib.classes
    model.collect_params().reset_ctx(ctx)
    return model

def resnet34_v1b_kinetics400(nclass=400, pretrained=False, tsn=False, partial_bn=False,
                             num_segments=1, root='~/.mxnet/models', ctx=mx.cpu(), **kwargs):
    if tsn:
        model = ActionRecResNetV1bTSN(depth=34,
                                      nclass=nclass,
                                      partial_bn=partial_bn,
                                      num_segments=num_segments)
    else:
        model = ActionRecResNetV1b(depth=34,
                                   nclass=nclass,
                                   partial_bn=partial_bn)

    if pretrained:
        from ..model_store import get_model_file
        model.load_parameters(get_model_file('resnet34_v1b_kinetics400',
                                             tag=pretrained, root=root))
        from ...data import Kinetics400Attr
        attrib = Kinetics400Attr()
        model.classes = attrib.classes
    model.collect_params().reset_ctx(ctx)
    return model

def resnet50_v1b_kinetics400(nclass=400, pretrained=False, tsn=False, partial_bn=False,
                             num_segments=1, root='~/.mxnet/models', ctx=mx.cpu(), **kwargs):
    if tsn:
        model = ActionRecResNetV1bTSN(depth=50,
                                      nclass=nclass,
                                      partial_bn=partial_bn,
                                      num_segments=num_segments)
    else:
        model = ActionRecResNetV1b(depth=50,
                                   nclass=nclass,
                                   partial_bn=partial_bn)

    if pretrained:
        from ..model_store import get_model_file
        model.load_parameters(get_model_file('resnet50_v1b_kinetics400',
                                             tag=pretrained, root=root))
        from ...data import Kinetics400Attr
        attrib = Kinetics400Attr()
        model.classes = attrib.classes
    model.collect_params().reset_ctx(ctx)
    return model

def resnet101_v1b_kinetics400(nclass=400, pretrained=False, tsn=False, partial_bn=False,
                              num_segments=1, root='~/.mxnet/models', ctx=mx.cpu(), **kwargs):
    if tsn:
        model = ActionRecResNetV1bTSN(depth=101,
                                      nclass=nclass,
                                      partial_bn=partial_bn,
                                      num_segments=num_segments)
    else:
        model = ActionRecResNetV1b(depth=101,
                                   nclass=nclass,
                                   partial_bn=partial_bn)

    if pretrained:
        from ..model_store import get_model_file
        model.load_parameters(get_model_file('resnet101_v1b_kinetics400',
                                             tag=pretrained, root=root))
        from ...data import Kinetics400Attr
        attrib = Kinetics400Attr()
        model.classes = attrib.classes
    model.collect_params().reset_ctx(ctx)
    return model

def resnet152_v1b_kinetics400(nclass=400, pretrained=False, tsn=False, partial_bn=False,
                              num_segments=1, root='~/.mxnet/models', ctx=mx.cpu(), **kwargs):
    if tsn:
        model = ActionRecResNetV1bTSN(depth=152,
                                      nclass=nclass,
                                      partial_bn=partial_bn,
                                      num_segments=num_segments)
    else:
        model = ActionRecResNetV1b(depth=152,
                                   nclass=nclass,
                                   partial_bn=partial_bn)

    if pretrained:
        from ..model_store import get_model_file
        model.load_parameters(get_model_file('resnet152_v1b_kinetics400',
                                             tag=pretrained, root=root))
        from ...data import Kinetics400Attr
        attrib = Kinetics400Attr()
        model.classes = attrib.classes
    model.collect_params().reset_ctx(ctx)
    return model



# pylint: disable=line-too-long
"""Encode/Decode helper function for compressed quantized models"""
import zlib
import base64

def encode_json(json_file, is_print=False):
    r""" Encode json string to compressed base64 string.

    Parameters
    ----------
    json_file : str
        String value represents the path to json file.
    is_print : bool
        Boolean value controls whether to print the encoded base64 string.
    """
    with open(json_file, encoding='utf-8') as fh:
        data = fh.read()
    zipped_str = zlib.compress(data.encode('utf-8'))
    b64_str = base64.b64encode(zipped_str)
    if is_print:
        print(b64_str)
    return b64_str

def decode_b64(b64_str, is_print=False):
    r""" Decode b64 string to json format

    Parameters
    ---------
    b64_str: str
        String value represents the compressed base64 string.
    is_print : bool
        Boolean value controls whether to print the decoded json string.
    """
    json_str = zlib.decompress(base64.b64decode(b64_str)).decode('utf-8')
    if is_print:
        print(json_str)
    return json_str

def get_compressed_model(model_name, compressed_json):
    r""" Get compressed (INT8) models from existing `compressed_json` dict

    Parameters
    ----------
    model_name: str
        String value represents the name of compressed (INT8) model.
    compressed_json : dict
        Dictionary's key represents the name of (INT8) model, and dictionary's value
        represents the compressed json string of (INT8) model.
    """
    b64_str = compressed_json.get(model_name, None)
    if b64_str:
        return decode_b64(b64_str)
    raise ValueError('Model: {} is not found. Available compressed models are:\n{}'.format(model_name, '\n'.join(list(compressed_json.keys()))))



"""Transforms for alpha pose estimation."""
from __future__ import absolute_import

import random
import numpy as np
import mxnet as mx
from ..pose import random_sample_bbox, count_visible, random_crop_bbox
from ..pose import drawGaussian, transformBox, cv_cropBox, cv_rotate, detector_to_alpha_pose


class AlphaPoseDefaultTrainTransform(object):
    """Default training transform for AlphaPose.

    Parameters
    ----------
    num_joints : int
        Number of joints defined by dataset
    joint_pairs: list
        Paired Joints for horizontal flipping
    image_size : tuple of int
        Image size, as (height, width).
    heatmap_size : tuple of int
        Heatmap size, as (height, width).
    sigma : float
        Gaussian sigma for the heatmap generation.
    random_flip: bool
        Data augmentation with horizaontal flipping.
    random_sample: bool
        Data augmentation with random sample.
    random_crop: bool
        Data augmentation with random crop.
    scale_factor: tuple
        Scale augmentation.
    rotation_factor: int
        Ratation augmentation.
    """
    def __init__(self, num_joints, joint_pairs, image_size=(256, 256), heatmap_size=(64, 64),
                 sigma=1, random_flip=True,
                 random_sample=False, random_crop=False,
                 scale_factor=(0.2, 0.3), rotation_factor=30):
        self._sigma = sigma
        self._num_joints = num_joints
        self._joint_pairs = joint_pairs
        self._image_size = image_size
        self._heatmap_size = heatmap_size
        self._height = image_size[0]
        self._width = image_size[1]
        self._random_flip = random_flip
        self._random_sample = random_sample
        self._random_crop = random_crop
        self._scale_factor = scale_factor
        self._rotation_factor = rotation_factor
        self._aspect_ratio = float(self._width) / self._height

    def __call__(self, src, label, img_path):
        src = src.asnumpy()
        bbox = label['bbox']
        assert len(bbox) == 4
        joints_3d = label['joints_3d']

        # color jitter src-(0, 255)
        src[:, :, 0] = (src[:, :, 0] * random.uniform(0.7, 1.3)).clip(0, 255)
        src[:, :, 1] = (src[:, :, 1] * random.uniform(0.7, 1.3)).clip(0, 255)
        src[:, :, 2] = (src[:, :, 2] * random.uniform(0.7, 1.3)).clip(0, 255)
        imght, imgwd = src.shape[0], src.shape[1]
        assert src.shape[2] == 3

        # scaling
        ul = np.array((int(bbox[0]), int(bbox[1])))
        br = np.array((int(bbox[2]), int(bbox[3])))
        h = br[1] - ul[1]
        w = br[0] - ul[0]
        sf = random.uniform(*self._scale_factor)
        ul[0] = max(0, ul[0] - w * sf / 2)
        ul[1] = max(0, ul[1] - h * sf / 2)
        br[0] = min(imgwd - 1, br[0] + w * sf / 2)
        br[1] = min(imght - 1, br[1] + h * sf / 2)
        if self._random_sample:
            ul, br = random_sample_bbox(ul, br, w, h, imgwd, imght)

        # boundary refine
        # ul, br = refine_bound(ul, br)

        # counting number of joints
        num_visible_joint, vis_joints = count_visible(ul, br, joints_3d)

        if self._random_crop and num_visible_joint > 10:
            ul, br = random_crop_bbox(ul, br)

        if num_visible_joint < 1:
            # no valid keypoints
            img = mx.nd.zeros((3, self._height, self._width))
            target = np.zeros((self._num_joints, self._heatmap_size[0], self._heatmap_size[1]),
                              dtype='float32')
            target_weight = np.ones((self._num_joints, 1, 1), dtype='float32')
            return img, target, target_weight, img_path

        # crop box with padding
        img = src
        img = cv_cropBox(img, ul, br, self._height, self._width)

        # generate labels
        target = np.zeros((self._num_joints, self._heatmap_size[0], self._heatmap_size[1]),
                          dtype='float32')
        target_weight = np.ones((self._num_joints, 1, 1), dtype='float32')

        for i, vis in enumerate(vis_joints):
            if vis:
                hm_part = transformBox(
                    (joints_3d[i, 0, 0], joints_3d[i, 1, 0]),
                    ul, br, self._height, self._width,
                    self._heatmap_size[0], self._heatmap_size[1])
                target[i] = drawGaussian(target[i], hm_part, self._sigma)
            target_weight[i] = 1

        # random flip
        if self._random_flip and random.random() > 0.5:
            assert img.shape[2] == 3
            img = img[:, ::-1, :]
            # target: (num_joints, h, w)
            target = flip_heatmap(target, self._joint_pairs)

        # rotation
        rf = self._rotation_factor
        r = np.clip(np.random.randn() * rf, -rf * 2, rf * 2) if random.random() <= 0.6 else 0
        img = cv_rotate(img, r, self._width, self._height)
        target = cv_rotate(target.transpose((1, 2, 0)), r,
                           self._heatmap_size[1], self._heatmap_size[0])
        target = target.transpose((2, 0, 1))
        # to tensor
        src = mx.nd.image.to_tensor(mx.nd.array(img))
        src[0] = src[0] - 0.406
        src[1] = src[1] - 0.457
        src[2] = src[2] - 0.480
        assert src.shape[0] == 3

        return src, target, target_weight, img_path


class AlphaPoseDefaultValTransform(object):
    """Default validation transform for AlphaPose.

    Parameters
    ----------
    num_joints : int
        Number of joints defined by dataset
    joint_pairs: list
        Paired Joints for horizontal flipping
    image_size : tuple of int
        Image size, as (height, width).
    sigma : float
        Gaussian sigma for the heatmap generation.

    """
    def __init__(self, num_joints, joint_pairs, image_size=(256, 256),
                 sigma=1):
        self._sigma = sigma
        self._num_joints = num_joints
        self._joint_pairs = joint_pairs
        self._image_size = image_size
        self._height = image_size[0]
        self._width = image_size[1]
        self._aspect_ratio = float(self._width) / self._height

    def __call__(self, src, label, img_path):
        src = src.asnumpy()
        bbox = label['bbox']
        assert len(bbox) == 4
        score = label.get('score', 1)
        img, scale_box = detector_to_alpha_pose(
            src,
            class_ids=mx.nd.array([[0.]]),
            scores=mx.nd.array([[1.]]),
            bounding_boxs=mx.nd.array(np.array([bbox])),
            output_shape=self._image_size)

        return img[0], scale_box, score, img_path


def flip_heatmap(heatmap, joint_pairs, shift=False):
    """Flip pose heatmap according to joint pairs.

    Parameters
    ----------
    heatmap : numpy.ndarray
        Heatmap of joints.
    joint_pairs : list
        List of joint pairs
    shift : bool
        Whether to shift the output

    Returns
    -------
    numpy.ndarray
        Flipped heatmap

    """
    assert heatmap.ndim == 3, "heatmap should have shape (num_joints, height, width)"
    out = heatmap[:, :, ::-1]

    for pair in joint_pairs:
        tmp = out[pair[0], :, :].copy()
        out[pair[0], :, :] = out[pair[1], :, :]
        out[pair[1], :, :] = tmp

    if shift:
        out[:, :, 1:] = out[:, :, 0:-1]
    return out



"""Mask Target Generator."""
from __future__ import absolute_import

from mxnet import gluon, autograd


class MaskTargetGenerator(gluon.HybridBlock):
    """Mask RCNN target encoder to generate mask targets.

    Parameters
    ----------
    num_images : int
        Number of input images.
    num_rois : int
        Number of sampled rois.
    num_classes : int
        Number of classes for class-specific targets.
    mask_size : tuple of int
        Size of generated masks, for example (14, 14).

    """

    def __init__(self, num_images, num_rois, num_classes, mask_size, **kwargs):
        super(MaskTargetGenerator, self).__init__(**kwargs)
        self._num_images = num_images
        self._num_rois = num_rois
        self._num_classes = num_classes
        self._mask_size = mask_size

    # pylint: disable=arguments-differ
    def hybrid_forward(self, F, rois, gt_masks, matches, cls_targets):
        """Handle B=self._num_image by a for loop.
        There is no way to know number of gt_masks.

        Parameters
        ----------
        rois: (B, N, 4), input proposals
        gt_masks: (B, M, H, W), input masks of full image size
        matches: (B, N), value [0, M), index to gt_label and gt_box.
        cls_targets: (B, N), value [0, num_class), excluding background class.

        Returns
        -------
        mask_targets: (B, N, C, MS, MS), sampled masks.
        box_target: (B, N, C, 4), only foreground class has nonzero target.
        box_weight: (B, N, C, 4), only foreground class has nonzero weight.

        """

        if hasattr(F.contrib, 'mrcnn_mask_target'):
            return F.contrib.mrcnn_mask_target(rois, gt_masks, matches, cls_targets,
                                               num_rois=self._num_rois,
                                               num_classes=self._num_classes,
                                               mask_size=self._mask_size)

        # cannot know M (num_gt) to have accurate batch id B * M, must split batch dim
        def _split(x, axis, num_outputs, squeeze_axis):
            x = F.split(x, axis=axis, num_outputs=num_outputs, squeeze_axis=squeeze_axis)
            if isinstance(x, list):
                return x
            elif self._num_images > 1:
                return list(x)
            else:
                return [x]

        with autograd.pause():
            # gt_masks (B, M, H, W) -> (B, M, 1, H, W) -> B * (M, 1, H, W)
            gt_masks = gt_masks.reshape((0, -4, -1, 1, 0, 0))
            gt_masks = _split(gt_masks, axis=0, num_outputs=self._num_images, squeeze_axis=True)
            # rois (B, N, 4) -> B * (N, 4)
            rois = _split(rois, axis=0, num_outputs=self._num_images, squeeze_axis=True)
            # remove possible -1 match
            matches = F.relu(matches)
            # matches (B, N) -> B * (N,)
            matches = _split(matches, axis=0, num_outputs=self._num_images, squeeze_axis=True)
            # cls_targets (B, N) -> B * (N,)
            cls_targets = _split(cls_targets, axis=0, num_outputs=self._num_images,
                                 squeeze_axis=True)

            # (1, C)
            cids = F.arange(1, self._num_classes + 1)
            cids = cids.reshape((1, -1))

            mask_targets = []
            mask_masks = []
            for roi, gt_mask, match, cls_target in zip(rois, gt_masks, matches, cls_targets):
                # batch id = match
                padded_rois = F.concat(match.reshape((-1, 1)), roi, dim=-1)
                # pooled_mask (N, 1, MS, MS)
                pooled_mask = F.contrib.ROIAlign(gt_mask, padded_rois,
                                                 self._mask_size, 1.0, sample_ratio=2)

                # (N,) -> (1, C) -> (N, C, 1, 1)
                cls_target = F.expand_dims(cls_target, 1)
                same_cids = F.broadcast_equal(cls_target, cids)
                same_cids = same_cids.reshape((-2, 1, 1))

                # (N, MS, MS) -> (N, C, 1, 1) -> (N, C, MS, MS)
                mask_mask = F.broadcast_like(same_cids, pooled_mask,
                                             lhs_axes=(2, 3), rhs_axes=(2, 3))

                # (N, 1, MS, MS) -> (N, C, MS, MS)
                mask_target = F.broadcast_axis(pooled_mask, size=self._num_classes, axis=1)

                mask_targets.append(mask_target)
                mask_masks.append(mask_mask)

            # B * (N, C, MS, MS) -> (B, N, C, MS, MS)
            mask_targets = F.stack(*mask_targets, axis=0)
            mask_masks = F.stack(*mask_masks, axis=0)
        return mask_targets, mask_masks



"""DeepLabV3+ with wideresnet backbone for semantic segmentation"""
# pylint: disable=missing-docstring,arguments-differ,unused-argument
from mxnet.gluon import nn
from mxnet.context import cpu
from mxnet.gluon.nn import HybridBlock
from .wideresnet import wider_resnet38_a2

__all__ = ['DeepLabWV3Plus', 'get_deeplabv3b_plus', 'get_deeplab_v3b_plus_wideresnet_citys']

class DeepLabWV3Plus(HybridBlock):
    r"""DeepLabWV3Plus

    Parameters
    ----------
    nclass : int
        Number of categories for the training dataset.
    backbone : string
        Pre-trained dilated backbone network type (default:'wideresnet').
    norm_layer : object
        Normalization layer used in backbone network (default: :class:`mxnet.gluon.nn.BatchNorm`;
        for Synchronized Cross-GPU BachNormalization).
    aux : bool
        Auxiliary loss.

    Reference:

        Chen, Liang-Chieh, et al. "Encoder-Decoder with Atrous Separable Convolution for Semantic
        Image Segmentation.", https://arxiv.org/abs/1802.02611, ECCV 2018
    """
    def __init__(self, nclass, backbone='wideresnet', aux=False, ctx=cpu(), pretrained_base=True,
                 height=None, width=None, base_size=520, crop_size=480, dilated=True, **kwargs):
        super(DeepLabWV3Plus, self).__init__()

        height = height if height is not None else crop_size
        width = width if width is not None else crop_size
        self._up_kwargs = {'height': height, 'width': width}
        self.base_size = base_size
        self.crop_size = crop_size
        print('self.crop_size', self.crop_size)

        with self.name_scope():
            pretrained = wider_resnet38_a2(classes=1000, dilation=True)
            pretrained.initialize(ctx=ctx)
            self.mod1 = pretrained.mod1
            self.mod2 = pretrained.mod2
            self.mod3 = pretrained.mod3
            self.mod4 = pretrained.mod4
            self.mod5 = pretrained.mod5
            self.mod6 = pretrained.mod6
            self.mod7 = pretrained.mod7
            self.pool2 = pretrained.pool2
            self.pool3 = pretrained.pool3
            del pretrained
            self.head = _DeepLabHead(nclass, height=height//2, width=width//2, **kwargs)
            self.head.initialize(ctx=ctx)

    def hybrid_forward(self, F, x):
        outputs = []
        x = self.mod1(x)
        m2 = self.mod2(self.pool2(x))
        x = self.mod3(self.pool3(m2))
        x = self.mod4(x)
        x = self.mod5(x)
        x = self.mod6(x)
        x = self.mod7(x)
        x = self.head(x, m2)
        x = F.contrib.BilinearResize2D(x, **self._up_kwargs)
        outputs.append(x)
        return tuple(outputs)

    def demo(self, x):
        return self.predict(x)

    def predict(self, x):
        h, w = x.shape[2:]
        self._up_kwargs['height'] = h
        self._up_kwargs['width'] = w
        x = self.mod1(x)
        m2 = self.mod2(self.pool2(x))
        x = self.mod3(self.pool3(m2))
        x = self.mod4(x)
        x = self.mod5(x)
        x = self.mod6(x)
        x = self.mod7(x)
        x = self.head.demo(x, m2)
        import mxnet.ndarray as F
        x = F.contrib.BilinearResize2D(x, **self._up_kwargs)
        return x

class _DeepLabHead(HybridBlock):
    def __init__(self, nclass, c1_channels=128, norm_layer=nn.BatchNorm, norm_kwargs=None,
                 height=240, width=240, **kwargs):
        super(_DeepLabHead, self).__init__()
        self._up_kwargs = {'height': height, 'width': width}
        with self.name_scope():
            self.aspp = _ASPP(in_channels=4096, atrous_rates=[12, 24, 36], norm_layer=norm_layer,
                              norm_kwargs=norm_kwargs, height=height//4, width=width//4, **kwargs)

            self.c1_block = nn.HybridSequential(prefix='bot_fine_')
            self.c1_block.add(nn.Conv2D(in_channels=c1_channels, channels=48,
                                        kernel_size=1, use_bias=False))

            self.block = nn.HybridSequential(prefix='final_')
            self.block.add(nn.Conv2D(in_channels=304, channels=256,
                                     kernel_size=3, padding=1, use_bias=False))
            self.block.add(norm_layer(in_channels=256,
                                      **({} if norm_kwargs is None else norm_kwargs)))
            self.block.add(nn.Activation('relu'))
            self.block.add(nn.Conv2D(in_channels=256, channels=256,
                                     kernel_size=3, padding=1, use_bias=False))
            self.block.add(norm_layer(in_channels=256,
                                      **({} if norm_kwargs is None else norm_kwargs)))
            self.block.add(nn.Activation('relu'))
            self.block.add(nn.Conv2D(in_channels=256, channels=nclass,
                                     kernel_size=1, use_bias=False))

    def hybrid_forward(self, F, x, c1):
        c1 = self.c1_block(c1)
        x = self.aspp(x)
        x = F.contrib.BilinearResize2D(x, **self._up_kwargs)
        return self.block(F.concat(c1, x, dim=1))

    def demo(self, x, c1):
        h, w = c1.shape[2:]
        self._up_kwargs['height'] = h
        self._up_kwargs['width'] = w
        c1 = self.c1_block(c1)
        x = self.aspp.demo(x)
        import mxnet.ndarray as F
        x = F.contrib.BilinearResize2D(x, **self._up_kwargs)
        return self.block(F.concat(c1, x, dim=1))

def _ASPPConv(in_channels, out_channels, atrous_rate, norm_layer, norm_kwargs):
    block = nn.HybridSequential()
    with block.name_scope():
        block.add(nn.Conv2D(in_channels=in_channels, channels=out_channels,
                            kernel_size=3, padding=atrous_rate,
                            dilation=atrous_rate, use_bias=False))
        block.add(norm_layer(in_channels=out_channels,
                             **({} if norm_kwargs is None else norm_kwargs)))
        block.add(nn.Activation('relu'))
    return block

class _AsppPooling(nn.HybridBlock):
    def __init__(self, in_channels, out_channels, norm_layer, norm_kwargs,
                 height=60, width=60, **kwargs):
        super(_AsppPooling, self).__init__()
        self.gap = nn.HybridSequential()
        self._up_kwargs = {'height': height, 'width': width}
        with self.gap.name_scope():
            self.gap.add(nn.GlobalAvgPool2D())
            self.gap.add(nn.Conv2D(in_channels=in_channels, channels=out_channels,
                                   kernel_size=1, use_bias=False))
            self.gap.add(norm_layer(in_channels=out_channels,
                                    **({} if norm_kwargs is None else norm_kwargs)))
            self.gap.add(nn.Activation("relu"))

    def hybrid_forward(self, F, x):
        pool = self.gap(x)
        return F.contrib.BilinearResize2D(pool, **self._up_kwargs)

    def demo(self, x):
        h, w = x.shape[2:]
        self._up_kwargs['height'] = h
        self._up_kwargs['width'] = w
        pool = self.gap(x)
        import mxnet.ndarray as F
        return F.contrib.BilinearResize2D(pool, **self._up_kwargs)

class _ASPP(nn.HybridBlock):
    def __init__(self, in_channels, atrous_rates, norm_layer, norm_kwargs,
                 height=60, width=60):
        super(_ASPP, self).__init__()
        out_channels = 256
        self.b0 = nn.HybridSequential()
        self.b0.add(nn.Conv2D(in_channels=in_channels, channels=out_channels,
                              kernel_size=1, use_bias=False))
        self.b0.add(norm_layer(in_channels=out_channels,
                               **({} if norm_kwargs is None else norm_kwargs)))
        self.b0.add(nn.Activation("relu"))

        rate1, rate2, rate3 = tuple(atrous_rates)
        self.b1 = _ASPPConv(in_channels, out_channels, rate1, norm_layer, norm_kwargs)
        self.b2 = _ASPPConv(in_channels, out_channels, rate2, norm_layer, norm_kwargs)
        self.b3 = _ASPPConv(in_channels, out_channels, rate3, norm_layer, norm_kwargs)
        self.b4 = _AsppPooling(in_channels, out_channels, norm_layer=norm_layer,
                               norm_kwargs=norm_kwargs, height=height, width=width)

        self.project = nn.HybridSequential(prefix='bot_aspp_')
        self.project.add(nn.Conv2D(in_channels=5*out_channels, channels=out_channels,
                                   kernel_size=1, use_bias=False))

    def hybrid_forward(self, F, x):
        feat1 = self.b0(x)
        feat2 = self.b1(x)
        feat3 = self.b2(x)
        feat4 = self.b3(x)
        x = self.b4(x)
        x = F.concat(x, feat1, feat2, feat3, feat4, dim=1)
        return self.project(x)

    def demo(self, x):
        feat1 = self.b0(x)
        feat2 = self.b1(x)
        feat3 = self.b2(x)
        feat4 = self.b3(x)
        x = self.b4.demo(x)
        import mxnet.ndarray as F
        x = F.concat(x, feat1, feat2, feat3, feat4, dim=1)
        return self.project(x)

def get_deeplabv3b_plus(dataset='citys', backbone='wideresnet', pretrained=False,
                        root='~/.mxnet/models', ctx=cpu(0), **kwargs):
    r"""DeepLabWV3Plus
    Parameters
    ----------
    dataset : str, default pascal_voc
        The dataset that model pretrained on. (pascal_voc, ade20k, citys)
    pretrained : bool or str
        Boolean value controls whether to load the default pretrained weights for model.
        String value represents the hashtag for a certain version of pretrained weights.
    ctx : Context, default CPU
        The context in which to load the pretrained weights.
    root : str, default '~/.mxnet/models'
        Location for keeping the model parameters.

    Examples
    --------
    >>> model = get_deeplabv3b_plus(dataset='citys', backbone='wideresnet', pretrained=False)
    >>> print(model)
    """
    acronyms = {
        'pascal_voc': 'voc',
        'pascal_aug': 'voc',
        'ade20k': 'ade',
        'coco': 'coco',
        'citys': 'citys',
    }
    from ..data import datasets
    # infer number of classes
    model = DeepLabWV3Plus(datasets[dataset].NUM_CLASS, backbone=backbone, ctx=ctx, **kwargs)
    model.classes = datasets[dataset].classes
    if pretrained:
        from .model_store import get_model_file
        model.load_parameters(get_model_file('deeplab_v3b_plus_%s_%s'%(backbone, acronyms[dataset]),
                                             tag=pretrained, root=root), ctx=ctx)
    return model

def get_deeplab_v3b_plus_wideresnet_citys(**kwargs):
    r"""DeepLabWV3Plus
    Parameters
    ----------
    pretrained : bool or str
        Boolean value controls whether to load the default pretrained weights for model.
        String value represents the hashtag for a certain version of pretrained weights.
    ctx : Context, default CPU
        The context in which to load the pretrained weights.
    root : str, default '~/.mxnet/models'
        Location for keeping the model parameters.

    Examples
    --------
    >>> model = get_deeplab_v3b_plus_wideresnet_citys(pretrained=True)
    >>> print(model)
    """
    return get_deeplabv3b_plus('citys', 'wideresnet', **kwargs)



"""RPN proposals."""
from __future__ import absolute_import

from mxnet import autograd
from mxnet import gluon

from ...nn.bbox import BBoxCornerToCenter, BBoxClipToImage
from ...nn.coder import NormalizedBoxCenterDecoder


class RPNProposal(gluon.HybridBlock):
    """Proposal generator for RPN.

    RPNProposal takes RPN anchors, RPN prediction scores and box regression predictions.
    It will transform anchors, apply NMS (if set to true) to get clean foreground proposals.

    Parameters
    ----------
    clip : float
        Clip bounding box target to this value.
    min_size : int
        Proposals whose size is smaller than ``min_size`` will be discarded.
    stds : tuple of float
        Standard deviation to be multiplied from encoded regression targets.
        These values must be the same as stds used in RPNTargetGenerator.
    nms : boolean
        Whether to do nms.
    """

    def __init__(self, clip, min_size, stds):
        super(RPNProposal, self).__init__()
        self._box_to_center = BBoxCornerToCenter()
        self._box_decoder = NormalizedBoxCenterDecoder(stds=stds, clip=clip, convert_anchor=True)
        self._clipper = BBoxClipToImage()
        # self._compute_area = BBoxArea()
        self._min_size = min_size

    # pylint: disable=arguments-differ
    def hybrid_forward(self, F, anchor, score, bbox_pred, img):
        """
        Generate proposals.
        """
        with autograd.pause():
            # restore bounding boxes
            roi = self._box_decoder(bbox_pred, anchor)

            # clip rois to image's boundary
            # roi = F.Custom(roi, img, op_type='bbox_clip_to_image')
            roi = self._clipper(roi, img)

            # remove bounding boxes that don't meet the min_size constraint
            # by setting them to (-1, -1, -1, -1)
            # width = roi.slice_axis(axis=-1, begin=2, end=3)
            # height = roi.slice_axis(axis=-1, begin=3, end=None)
            xmin, ymin, xmax, ymax = roi.split(axis=-1, num_outputs=4)
            width = xmax - xmin + 1.0
            height = ymax - ymin + 1.0
            # TODO:(zhreshold), there's im_ratio to handle here, but it requires
            # add' info, and we don't expect big difference
            invalid = (width < self._min_size) + (height < self._min_size)

            # # remove out of bound anchors
            # axmin, aymin, axmax, aymax = F.split(anchor, axis=-1, num_outputs=4)
            # # it's a bit tricky to get right/bottom boundary in hybridblock
            # wrange = F.arange(0, 2560).reshape((1, 1, 1, 2560)).slice_like(
            #    img, axes=(3)).max().reshape((1, 1, 1))
            # hrange = F.arange(0, 2560).reshape((1, 1, 2560, 1)).slice_like(
            #    img, axes=(2)).max().reshape((1, 1, 1))
            # invalid = (axmin < 0) + (aymin < 0) + F.broadcast_greater(axmax, wrange) + \
            #    F.broadcast_greater(aymax, hrange)
            # avoid invalid anchors suppress anchors with 0 confidence
            score = F.where(invalid, F.ones_like(invalid) * -1, score)
            invalid = F.broadcast_axes(invalid, axis=2, size=4)
            roi = F.where(invalid, F.ones_like(invalid) * -1, roi)

            pre = F.concat(score, roi, dim=-1)
            return pre
